
class CrewSizeWarning(UserWarning):
    base_msg = 'Ignition crew size for this firing technique should be ' \
        'exactly'
    only_using_one = f'{base_msg} 1. Only the first igniter in the crew will be used'
    only_using_two = f'{base_msg} 2. Only the first two igniters in the crew will be used'
    cloning_first = f'{base_msg} 2. The first igniter was dulicated to increase ' \
        'the crew to the minimum required size.'



"""
Reading and manipulating a burn unit geometry and splitting
the boundary to wind-centric segments
"""

# Core imports
from __future__ import annotations
import copy

# Internal imports
from driptorch.io import Projector, write_geojson, read_geojson_polygon

# External imports
import numpy as np
from shapely.geometry import Polygon, LineString, Point
from shapely import affinity


# TODO: #8 Base class shapely polygon and extend in BurnUnit class
class BurnUnit:

    def __init__(self, polygon: Polygon, wind_direction: float, utm_epsg: int = None):
        """Constructor

        Args:
            polygon (Polygon): Shapely polygon geometry object
            wind_direction (float): Wind direction (degrees)
            utm (int): UTM EPSG code of spatial data. If not provided it is assumed
                that coordinates are in 4326 and will be converted to UTM. Defaults to None.
        """

        # Set the global EPSG source
        if not utm_epsg:
            utm_epsg, polygon = Projector.web_mercator_to_utm(polygon)

        # Store instance attributes
        self.utm_epsg = utm_epsg
        self.polygon = polygon
        self.wind_direction = wind_direction
        self.centroid = polygon.centroid
        self.polygon_segments = PolygonSplitter()

        # Compute the angle used to rotate the unit s.t. the wind direction
        # is congruent with the positive x-axis
        self.wind_alignment_angle = (self.wind_direction - 90) % 360

        # Align the unit, run the boundary splitting algorithm for
        # anchors and segs, and then unalign
        self._align()
        self.polygon_segments.split(self.polygon)
        self._unalign()

    @classmethod
    def from_json(cls, geojson: dict, wind_direction: float) -> BurnUnit:
        """Alternative constructor used to create a BurnUnit object from
        a GeoJSON dictionary.

        Args:
            geojson (dict): GeoJSON dictionary
            wind_direction (float): Wind direction (degrees)
            epsg (int): EPSG code of the input GeoJSON. Defaults to 4326.

        Returns:
            BurnUnit: New instance of a BurnUnit
        """

        # Read the GeoJSON as a shapely polygon
        polygon = read_geojson_polygon(geojson)

        return cls(polygon, wind_direction)

    def to_json(self, **kwargs) -> dict:
        """Write the BurnUnit boundary to a GeoJSON dictionary

        Returns:
            dict: GeoJSON dictionary
        """

        return write_geojson([self.polygon], utm_epsg=self.utm_epsg, **kwargs)

    def _align(self):
        """Align the unit and boundary segs to the wind
        """

        if self.wind_alignment_angle:
            self.polygon = affinity.rotate(
                self.polygon, self.wind_alignment_angle, self.centroid)
            self.polygon_segments.rotate(
                self.wind_alignment_angle, self.centroid)

    def _unalign(self):
        """Revert the unit and boundary seg to their origional orientation
        """

        if self.wind_alignment_angle:
            self.polygon = affinity.rotate(
                self.polygon, -self.wind_alignment_angle, self.centroid)
            self.polygon_segments.rotate(
                -self.wind_alignment_angle, self.centroid)

    def copy(self) -> BurnUnit:
        """Utility method for copying a BurnUnit instance

        Returns:
            BurnUnit: New instance of a BurnUnit
        """

        return copy.copy(self)

    def buffer_control_line(self, width: float) -> BurnUnit:
        """Shrink the burn unit to account for the width of the control line

        Args:
            distance (with): Width of the control line (meters)

        Returns:
            BurnUnit: New instance of a BurnUnit
        """

        # Use shapely's buffer method on the polygon
        buffered_polygon = self.polygon.buffer(-width)

        return BurnUnit(buffered_polygon, self.wind_direction, utm_epsg=self.utm_epsg)

    def buffer_downwind(self, width: float) -> BurnUnit:
        """Create a downwind blackline buffer

        Args:
            width (float): Width of downwind buffer (meters)

        Returns:
            BurnUnit: New instance of a BurnUnit
        """

        # Buffer the fore line by the blackline width parameter
        fore_line_buffer = self.polygon_segments.fore.buffer(
            width, cap_style=3)

        # Now take the difference between the current firing area and the fore line
        # buffer to cut out the downwind blackline area
        buffered_polygon = self.polygon.difference(fore_line_buffer)

        return BurnUnit(buffered_polygon, self.wind_direction, utm_epsg=self.utm_epsg)

    def difference(self, burn_unit: BurnUnit) -> BurnUnit:
        """Return a burn unit instance that is the difference between
        this burn unit and another. Useful for obtaining the blackline
        area for clearing fuels.

        Args:
            burn_unit (BurnUnit): The BurnUnit instance to difference against

        Returns:
            BurnUnit: A new instance of a BurnUnit
        """

        # Use shapely's set-theorectic difference method
        polygon_difference = self.polygon.difference(burn_unit.polygon)

        return BurnUnit(polygon_difference, self.wind_direction, utm_epsg=self.utm_epsg)

    def get_bounds(self) -> np.ndarray:
        """Helper method for fetching the bounding box of the unit

        Returns:
            np.ndarray: Bounding box of the unit
        """

        return np.array(self.polygon.envelope.exterior.coords[:-1])


class PolygonSplitter:

    def __init__(self):
        """Constructor
        """

        # We're just setting these to None to start with so they can be pumped
        # through wind alignment before we compute them
        self.fore: LineString = None
        self.aft: LineString = None
        self.port: LineString = None
        self.starboard: LineString = None

    def rotate(self, angle: float, origin: Point):
        """Helper method to rotate all four boundary segs

        Args:
            angle (float): Angle to rotate line by (degrees)
            origin (Point): Origin of rotation
        """

        self.fore = affinity.rotate(
            self.fore, angle, origin) if self.fore else self.fore
        self.aft = affinity.rotate(
            self.aft, angle, origin) if self.aft else self.aft
        self.port = affinity.rotate(
            self.port, angle, origin) if self.port else self.port
        self.starboard = affinity.rotate(
            self.starboard, angle, origin) if self.starboard else self.starboard

    def split(self, polygon: Polygon):
        """Split the polygon to four wind-centric segments

        Args:
            polygon (Polygon): Polygon to split
        """

        self.coords = np.array(polygon.exterior.coords[:-1])

        # Extract fore, aft, port and starboard anchor points
        fore_idx = self._get_anchor(0)
        aft_idx = self._get_anchor(0, side='upper')
        port_idx = self._get_anchor(1)
        starboard_idx = self._get_anchor(1, side='upper')

        # Extract fore, aft, port and starboard lines
        self.fore = self._get_segment(port_idx, starboard_idx)
        self.aft = self._get_segment(starboard_idx, port_idx)
        self.port = self._get_segment(aft_idx, fore_idx)
        self.starboard = self._get_segment(fore_idx, aft_idx)

    def _get_segment(self, start_idx: int, end_idx: int) -> LineString:
        """Compute burn unit exterior line strings (subsequence of polygon coordinates)
        that make up the four wind-centric parts of the burn unit: fore, aft, port and starboard.

        Args:
            start_idx (int): Starting index in coordinate sequence for line
            end_idx (int): Ending index in coordinate sequence for line

        Returns:
            LineString: wind-centric subsequence of the polygon
        """

        # If the start index is greater than the end index, then we need to roll
        # through the end of the array and concatenate the to segments
        if start_idx > end_idx:
            seg_a = self.coords[start_idx:]
            seg_b = self.coords[:end_idx+1]
            line = np.concatenate([seg_a, seg_b])

        # Otherwise, we can grab the entire substring in one go
        else:
            line = self.coords[start_idx:end_idx+1]

        # Convert to shapely LineString object and return
        return LineString(line)

    def _get_anchor(self, dimension: int, side='lower') -> int:
        """Compute anchor point indecies

        Args:
            dimension (int): Dimension to extract anchor from (x or y)
            side (str, optional): Lower or upper side of the burn unit. Defaults to 'lower'.

        Returns:
            int: Index of anchor point in the polygon coordinate sequence
        """

        # Condtionally extract minimum and maximum anchor point in the linear
        # ring for the specified dimension
        if side == 'lower':
            idx = np.where(self.coords[:, dimension] ==
                           self.coords[:, dimension].min())[0]
        else:
            idx = np.where(self.coords[:, dimension] ==
                           self.coords[:, dimension].max())[0]

        # If more than one anchor, select the most upwind anchor
        if len(idx) > 1:
            idx = idx[np.argmax(self.coords[idx, 0])]
        else:
            idx = idx[0]

        # Convert to shapely Point geometry and return
        return idx



"""
Spatiotemporal patterns and the infamous temporal propagator
"""

# Core imports
from __future__ import annotations
from time import time as unix_time

# Internal imports
from shapely.errors import ShapelyDeprecationWarning
from shapely import affinity
import warnings
from driptorch.io import write_geojson, write_quicfire

# External imports
import awkward as ak
import numpy as np
import pandas as pd
from shapely.geometry import MultiPoint, MultiLineString, LineString

# Turn off Pandas copy warning (or figure out how to do it like the Panda wants)
pd.options.mode.chained_assignment = None

# Turn off the Shapely deprecation warning about about future removal
# of the array interface. This happens when a Pandas takes a list of Shapely
# geometies and casts to a list of ndarray objects. GeoPandas would fix this
# but it's not worth the added complexity of GDAL C binary deps. This is fine
# as long as we don't upgrade the Shapely req to v2.
warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)


class Pattern:

    def __init__(self, heat: list[int], igniter: list[int], leg: list[int],
                 times: list[list[float]], geometry: list[LineString], utm_epsg: int):
        """Constructor

        Args:
            heat (list[int]): Heat of the path
            igniter (list[int]): Igniter id assigned to the path
            leg (list[int]): Leg for the path
            times (list[list[float]]): Coordinate arrival times for the path
            geometry (list[LineString]): Path geometry
            utm_epsg (int): UTM EPSG code for the CRS that the paths are projected in.
        """

        self.heat = heat
        self.igniter = igniter
        self.leg = leg
        self.times = times
        self.geometry = geometry
        self.utm_epsg = utm_epsg

    @classmethod
    def from_dict(cls, paths_dict: dict, utm_epsg: int) -> Pattern:
        """Alternative contructor for initializing a Pattern object with a dictionary
        of path parameters

        Args:
            paths_dict (dict): Dictionary of path parameters
            utm_epsg (int): UTM EPSG code for the CRS that the paths are currently projected in.

        Returns:
            Pattern: A new instance of Pattern
        """

        return cls(
            paths_dict['heat'],
            paths_dict['igniter'],
            paths_dict['leg'],
            paths_dict['times'],
            paths_dict['geometry'],
            utm_epsg=utm_epsg
        )

    def to_dict(self) -> dict:
        """Returns the Pattern path parameters as a dictionary

        Returns:
            dict: Pattern path parameters
        """

        return {
            'heat': self.heat,
            'igniter': self.igniter,
            'leg': self.leg,
            'times': self.times,
            'geometry': self.geometry
        }

    @staticmethod
    def empty_path_dict() -> dict:
        """Helper method for initializing a path parameter dictionary

        Returns:
            dict: Empty path dictionary
        """

        return {
            'heat': [],
            'igniter': [],
            'leg': [],
            'geometry': []
        }

    def to_json(self) -> dict:
        """Write the Pattern to a GeoJSON dictionary

        Returns:
            dict: Timestamped GeoJSON representation of the firing pattern
        """

        # The Timedstamed GeoJSON plug in won't take a time for each coordinate in
        # the sub line string of a MLS, apparently it wants a single time to represent
        # the entire sub line (Either they have a bug or I'm missing something).
        for i, geom in enumerate(self.geometry):
            if isinstance(geom, MultiLineString):
                # Only keep the start time for each sub line
                self.times[i] = [time[0] for time in self.times[i]]

        # Read the jagged times array as an Awkward array for vectorized operations
        times = ak.Array(self.times)

        # Convert to milliseconds since Epoch (this is what Leaflet wants)
        times = (times * 1000) + (unix_time() * 1000)
        self.times = times.to_list()

        # Set the props and styling
        props = {'heat': self.heat, 'igniter': self.igniter,
                 'leg': self.leg, 'times': self.times}
        style = {'icon': 'circle', 'style': {
            'color': '#ff0000', 'radius': 1}}

        # Send off to the GeoJSON writer and return
        return write_geojson(self.geometry, self.utm_epsg, properties=props, style=style)

    def translate(self, x_off: float, y_off: float) -> Pattern:

        geoms = self.geometry
        trans_geoms = []
        for geom in geoms:
            trans_geoms.append(affinity.translate(geom, x_off, y_off))

        self.geometry = trans_geoms

        return self

    def to_quicfire(self, filename: str = None) -> None | str:
        """Write paths dictionary to QUIC-fire ignition file format.

        Args:
            filename (str, optional): If provided, write the ignition file to the
                filename. Defaults to None.

        Returns:
            None | str: None if filename provided, string containing the ignition file if not.
        """

        if filename:
            with open(filename, 'w') as f:
                f.write(write_quicfire(self.geometry, self.times))
        else:
            return write_quicfire(self.geometry, self.times)


class TemporalPropagator:
    """
    This class takes spatial ignition paths and propagates time
    through their coordinates.
    """

    def __init__(self, spacing: float, sync_end_time: bool = False):
        """Class constructor

        Args:
            spacing (float): Stagger spacing between igniters in a heat (meters)
            sync_end_time (bool, optional): If true synchronize igniters within
                a heat to finish simultaneously. Defaults to False.
        """

        self.spacing = spacing
        self.sync_end_time = sync_end_time

    def forward(self, paths, ignition_crew):

        # Create a Pandas DataFrame from the initialized paths dictionary
        self.paths = pd.DataFrame(paths)
        self.ignition_crew = ignition_crew

        # TODO: #6 Check for geometry, heat, igniter and leg columns in Propagator.
        # Geometry must of type LineString

        # Setup some new dataframe columns
        self.paths['start_time'] = 0
        self.paths['end_time'] = 0
        self.paths['times'] = None
        self.paths['times'] = self.paths['times'].astype('object')

        # Sort dataframe by heat, igniter, leg (in that order)
        self.paths.sort_values(by=['heat', 'igniter', 'leg'], ascending=[
            True, True, True], inplace=True)

        # Run the initial forward pass through the paths
        self._init_path_time(self.spacing)

        # Synchronize within heat end times if specified (e.g. ring ignition)
        if self.sync_end_time:
            self._sync_end_time()

        # Compute the arrival time for each coordinate in each path
        self._compute_arrival_times()

        # Drop the intermidary columns
        self.paths.drop(['start_time', 'end_time'],
                        axis=1, inplace=True)

        return self.paths.to_dict(orient='list')

    def _init_path_time(self, spacing: float):
        """ Helper method to run the initial time propagation.

        Args:
            spacing (float): Stagger spacing between igniters in a heat
        """

        # Loop over each ignition path and compute the start time and end time
        for index, path in self.paths.iterrows():

            # Get the heat, igniter and leg indices from the current path
            i, j, k = path.heat, path.igniter, path.leg
            velocity = self.ignition_crew[j].velocity

            # The first check here handles multileg paths. If we're not on the
            # first leg, then we need to compute the distance between the last
            # coordinate of the previous leg and the first coordinate of the
            # current leg and see how long it takes the igniter to get there
            if k != 0:

                # Get the previous path leg
                prev_leg = self.paths.loc[(self.paths.heat == i) & (
                    self.paths.igniter == j) & (self.paths.leg == k-1)]

                # Grab the geometries of the previous and current leg
                prev_leg_geom = prev_leg.geometry.iloc[0]
                cur_leg_geom = path.geometry

                # The shapely distance function gives the shortest distance between
                # the geometries
                distance = prev_leg_geom.distance(cur_leg_geom)

                # Now take the end time of the previous leg and the travel time
                # to get the start time of the current leg
                prev_leg_end_time = prev_leg.end_time.values[0]
                path.start_time = prev_leg_end_time + distance / velocity

            # This check is for the first igniter of a heat (except the first heat,
            # that igniter's start is time zero). The first igniter of a heat can
            # start after the maximum end time of all igniters in the previous heat
            elif (i != 0) and (j == 0):

                # Get the maximum end time from all igniters in the previous heat
                prev_heat_max_end_time = self.paths.loc[self.paths.heat ==
                                                        i-1, 'end_time'].max()

                # The current igniter's start time is the previous heat max end time
                path.start_time = prev_heat_max_end_time

            # This check is for any igniter following the first one. Note that
            # multileg ignition paths are caught in the first check above, so
            # this always deals with the first leg. We'll start by giving the
            # igniter the same start time as the previous one, but igniters don't
            # always start along a line parallel to a coordinate frame axis and
            # ignition paths are not always straight lines. So, we have to do a
            # vector projection to compute the offset in the helper function
            # `_get_offset`. We'll also apply the spacing there as well.
            elif j != 0:

                # Get the previous and current igniters (again, no need to use
                # k the leg index, these will all be leg 0 due to the catch
                # above)
                cur_igniter = self.paths.loc[(
                    self.paths.heat == i) & (self.paths.igniter == j)]
                prev_igniter = self.paths.loc[(
                    self.paths.heat == i) & (self.paths.igniter == j-1)]

                # Get the previous igniter's start time
                prev_start_time = prev_igniter.start_time.values[0]

                # Compute the offset due to stagger spacing and incongruence
                # along a coordinate frame axis
                path.start_time = prev_start_time + self._get_offset(
                    prev_igniter, cur_igniter, spacing, velocity)

            # The last condition we have caught yet is the first igniter of the
            # first heat. That igniter gets a start time of zero
            else:
                path.start_time = 0

            # Calculate path end time (seconds along path plus start time)
            # path.end_time = path.start_time + path.meters / velocity
            path.end_time = path.start_time + path.geometry.length / velocity

            # Insert row back into paths dataframe
            self.paths.loc[index] = path

        # Depending of the arrange of igniter start positions it is
        # possible that the minimum start time is less than zero. We fix
        # that by adding the minimum back to all start and end times
        min_start_time = self.paths['start_time'].min()
        if min_start_time != 0:
            self.paths['start_time'] -= min_start_time
            self.paths['end_time'] -= min_start_time

    def _get_offset(self, prev_igniter: pd.Series, cur_igniter: pd.Series, spacing: float, velocity: float) -> float:
        """Compute the offset time between the previous and current igniters

        Args:
            prev_igniter (pd.Series): Dataframe row of previous igniter
            cur_igniter (pd.Series): Dataframe row of current igniter
            spacing (float): Stagger spacing (meters)
            velocity (float): Speed of the igniter (meters/second)

        Returns:
            float: Offset time between previous and current igniters
        """

        # We need three coordinates to construct two vectors: the
        # first coordinate from the current igniter and the first
        # and second coordinate from the preivous igniter
        cur_igniter_first_pos = np.array(
            cur_igniter.geometry.iloc[0].coords[0])
        prev_igniter_first_pos = np.array(
            prev_igniter.geometry.iloc[0].coords[0])
        prev_igniter_second_pos = np.array(
            prev_igniter.geometry.iloc[0].coords[1])

        # Now construct two vectors. First a vector from the
        # previous igniter's first position to the current igniter's
        # first position. And second a vector from the previous
        # igniters first position to the previous igniters second
        # position
        a_vec = cur_igniter_first_pos - prev_igniter_first_pos
        b_vec = prev_igniter_second_pos - prev_igniter_first_pos

        # Now we can project the a vector onto the unit b vector and
        # compute its magnitude which give the offset distance
        offset_distance = a_vec.dot(b_vec/np.linalg.norm(b_vec))

        # Add the stagger spacing distance and divide by velocity
        # for travel time
        return (spacing + offset_distance) / velocity

    def _sync_end_time(self):
        """ Helper method to synchronize end times"""

        # TODO: #7 What about multileg ignition paths when synchronizing end times?

        # Get the unique heat indecies
        heats = np.sort(self.paths.heat.unique())

        # For each heat find the maximum end time and adjust the start time accordingly
        for i in heats:

            # Grab the dataframe rows for the current heat
            cur_heat = self.paths[self.paths.heat == i]

            # Compute difference between start time and max end time
            max_end_time = cur_heat.end_time.max()
            time_diff = max_end_time - cur_heat.end_time

            # Adjust the start time s.t. end time is equal to the max end time
            cur_heat.start_time = cur_heat.start_time + time_diff
            cur_heat.end_time = max_end_time

            # Place the current heat back in the dataframe
            self.paths[self.paths.heat == i] = cur_heat

    def _compute_arrival_times(self):
        """For each path, ask the igniter what line type they are
        suppose to lay down and call the respective methods.
        """

        for index, path in self.paths.iterrows():
            cur_igniter = self.ignition_crew[path.igniter]
            if cur_igniter.interval == 0:
                self._lines(index, path, cur_igniter.velocity)
            elif cur_igniter.interval < 0:
                self._dashes(index, path, cur_igniter.velocity,
                             -cur_igniter.interval)
            else:
                self._dots(index, path, cur_igniter.velocity,
                           cur_igniter.interval)

    def _lines(self, index: int, path: pd.Series, velocity: float):
        """Compute arrival times along each igniter's coordinate sequence
        without stopping (solid ignition paths)
        """

        # Get the coordinate sequence as a numpy array
        coords = np.array(path.geometry.coords)

        # Reset the arrival time to the start time of the current path
        arrival_time = path.start_time

        # Initialize the arrival time array for the current path
        path_times = [arrival_time]

        # Loop of the coordinate sequence except the last one
        for i, xy in enumerate(coords[:-1]):

            # Get the next coordinate in the sequence and compute the distance
            # between it and the current coordinate
            next_xy = coords[i+1]
            meters = np.linalg.norm(xy - next_xy)

            # Update the arrival time and add to the times array for the
            # current path
            arrival_time = arrival_time + (meters/velocity)
            path_times.append(arrival_time)

        self.paths.at[index, 'times'] = path_times

    def _dashes(self, index: int, path: pd.Series, velocity: float, interval: float):
        """Compute arrival times along each igniter's coordinate sequence
        by picking up and putting down ignitions (dashes)
        """

        # We'll have a start time and end time for each dash. Initialize
        # the start time to the current path's start time
        start_time = path.start_time

        # Get a range of distances along the length of the path spaced by
        # the ignition rate
        distances = np.arange(0, path.geometry.length, interval)

        # Interpolate points at those distance along the path
        points = MultiPoint([path.geometry.interpolate(distance)
                             for distance in distances])

        # Set the arrays for the current path arrival times and
        # dash geometries
        path_times = []
        line_segs = []

        # Reset the fire boolean for controlling when the ignition is active
        fire = True
        point_geoms = list(points.geoms)

        # Loop over the interpolated points except the last
        for i, point in enumerate(point_geoms[:-1]):

            # Get the current and next point coordinates
            xy = np.array(point.coords)[0]
            next_xy = np.array(point_geoms[i+1].coords)[0]

            # Comput the distance between points and the travel time
            meters = np.linalg.norm(xy - next_xy)
            end_time = float(start_time + (meters/velocity))

            # IF fire is active for this segment then add start and end
            # times to the current path arrival time arrray and append
            # the dash to the line segment array
            if fire:
                path_times.append([start_time, end_time])
                # path_times.append(end_time)
                line_segs.append([xy.tolist(), next_xy.tolist()])

            # Toogle the fire boolean and update the start time
            fire ^= True
            start_time = end_time

        # Add the current path arrival times to the global times array
        # and create a multileg line string and append to the new geometry
        # array
        self.paths.at[index, 'times'] = path_times
        self.paths.at[index, 'geometry'] = MultiLineString(line_segs)

    def _dots(self, index: int, path: pd.Series, velocity: float, interval: float):
        """Compute arrival times along each igniter's coordinate sequence
        for interpolated equally spaced points (dots)
        """

        arrival_time = path.start_time

        # Get a range of distances along the length of the path space by
        # the ignition rate
        distances = np.arange(0, path.geometry.length, interval)

        # Interpolate points at those distance along the path
        points = MultiPoint([path.geometry.interpolate(distance)
                             for distance in distances])

        # Set the initial arrival time
        path_times = [arrival_time]

        # For each interpolated point except the last
        for i, point in enumerate(points[:-1]):

            # Get the current and next points
            xy = np.array(point.coords)
            next_xy = np.array(points[i+1].coords)

            # Compute the distance between points and travel time
            meters = np.linalg.norm(xy - next_xy)
            arrival_time = arrival_time + (meters/velocity)

            # Add the arrival time to the current path's time array
            path_times.append(arrival_time)

        # Add the current path's time array to the global time array and
        # add the multipar point geometry to the new geometries array
        self.paths.at[index, 'times'] = path_times
        self.paths.at[index, 'geometry'] = points



"""
Burn operation personnel
"""

# Core imports
from __future__ import annotations
import copy

# Internal imports
from driptorch.errors import *


class Igniter:
    """An igniter is anything with a velocity and ignition interval, like a person
    carrying a drip torch or a drone dispatching DAIDs.
    """

    # TODO: #5 Write serialization method for Igniter class
    def __init__(self, velocity: float, rate: float, rate_units: str = 'meters'):
        """Constructor computes and stores the interval in ignitions per meter (ipm)
        and ignitions per second (ips).

        Args:
            velocity (float): Speed of the igniter (meters/second)
            rate (float): Ignition rate in ipm (ignitions per meter) or ips (ignitions per second).
                Use the `rate_units` parameter to specifiy meters or seconds. An interval of 0 specifies
                a solid ignition line, while a negative value denotes a dashed ignition line and positve a
                dotted ignition line.
            rate_units (str, optional): Units for the ignition rate, must be "meters" or "seconds".
                Defaults to 'meters'.
        """

        self.velocity = velocity

        # Solid ignition line
        if rate == 0:
            self.interval = 0

        # Compute ipm and ips for dashes and dots
        elif rate_units == 'seconds':
            self.interval = 1.0 / (rate / velocity)
        else:
            self.interval = 1.0/rate

    def copy(self) -> Igniter:
        """Sometimes we need to copy a particular Igniter because they're so good
        at what they do.

        Returns:
            driptorch.Igniter: A copy of the this Ignitor object
        """

        return copy.copy(self)

    def __str__(self):

        return (f'Igniter(velocity={self.velocity}, ' +
                f'interval_ipm={self.interval_ipm}, ' +
                f'interval_ips={self.interval_ips})')


class IgnitionCrew:
    """
    An ignition crew is a collection of igniters. Sometime you may want your igniters
    to all have the same velocity and/or interval. You can specifiy these constraints
    in the constructor.
    """

    # TODO: #2 Include same_interval boolean in constructor of IgnitionCrew and implement validator
    # TODO: #3 Write serialization method for IgnitionCrew object
    # TODO: #4 Improve IgnitionCrew __str__() method

    def __init__(self, same_velocity: bool = True, same_rate: bool = True):
        """Constructor

        Args:
            same_velocity (bool, optional): True requires all igniters of an instance
                to have equal velocities. Defaults to True.
            same_rate (bool, optional): True requires all igniter of an instance
                to have equal rate values. Defaults to True.
        """

        self._same_velocity = same_velocity
        self._same_rate = same_rate
        self._velocity_req = None
        self._rate_req = None

        self._igniters = []

    @classmethod
    def from_list(cls, igniters: list[Igniter], **kwargs) -> IgnitionCrew:
        """Alternate constructor for building an ignition crew from a list of igniters

        Args:
            igniters (List[Igniter]): List of Igniter objects

        Returns:
            IgnitionCrew: An IgnitionCrew object with igniters from provided list
        """

        ignition_crew = cls(**kwargs)

        # Add igniters from provided list to the crew object
        for igniter in igniters:
            ignition_crew.add_igniter(igniter)

        return ignition_crew

    @classmethod
    def clone_igniter(cls, igniter: Igniter, clones: int, **kwargs) -> IgnitionCrew:
        """Alternate constructor for building an ignition crew by cloning a given
        igniters `n` times.

        Args:
            igniter (Igniter): The Igniter object to clone
            clones (int): Number of clones (number of igniters in crew)

        Returns:
            IgnitionCrew: An IgnitionCrew object with `n` clones of the specified Igniter
        """

        igniters = [igniter.copy() for _ in range(clones)]

        return cls.from_list(igniters, **kwargs)

    def add_igniter(self, igniter: Igniter):
        """Add an igniter to the crew

        Args:
            igniter (Igniter): Igniter object to add to the crew
        """

        # Check the igniter's velocity
        self._validate_velocity(igniter.velocity)
        self._validate_rate(igniter.interval)

        # If the validator didn't raise an exception, then add the igniter to the crew
        self._igniters.append(igniter)

    def _validate_velocity(self, velocity: float):
        """Private helper method to validate the velcity of the candidate igniter
        against the velocity requirement of the crew.

        Args:
            velocity (float): Velocity of the candidate igniter

        Raises:
            IgniterError: Exception raised if igniter's velocity is invalid
        """

        if self._same_velocity:
            if self._velocity_req:
                if velocity != self._velocity_req:
                    raise IgniterError(IgniterError.unequal_velocities)
            else:
                self._velocity_req = velocity

    def _validate_rate(self, rate: float):
        """Private helper method to validate the rate of the candidate igniter
        against the rate requirement of the crew.

        Args:
            rate (float): Ignition rate of the candidate igniter

        Raises:
            IgniterError: Exception raised if igniter's rate is invalid
        """

        if self._same_rate:
            if self._rate_req:
                if rate != self._rate_req:
                    raise IgniterError(IgniterError.unequal_rates)
            else:
                self._rate_req = rate

    def __getitem__(self, index):

        return self._igniters[index]

    def __len__(self):

        return len(self._igniters)

    def __iter__(self):

        for i in range(self.__len__()):
            yield self._igniters[i]

    def __str__(self):

        print_str = f'IgnitionCrew\n'
        for igniter in self._igniters:
            print_str += f'{igniter}\n'
        return print_str




class GeojsonError(Exception):
    read_error = 'Invalid GeoJSON format, that is all we know'


class IgniterError(Exception):
    unequal_velocities = 'Cannot add igniter to crew due to unequal velocities. ' \
        'You can set `same_velocity=False` in the constructor, ' \
        'however this can raise subsequent errors if an ignition ' \
        'crew with unequal velocity is passed to an internal ignition ' \
        'pattern generation method.'
    unequal_rates = 'Cannot add igniter to crew due to unequal rates. ' \
        'You can set `same_rate=False` in the constructor, ' \
        'however this can raise subsequent errors if an ignition ' \
        'crew with unequal rates is passed to a fire module exporter.'


class ExportError(Exception):
    incompatible_line_types = 'QUIC-fire exports can not include ' \
        'Point and LineString/MultiLineString geometry types in the ' \
        'same file. Please reconfigure your ignition crew to have either ' \
        'all point ignitions or line/dash ignitions.'




from driptorch.unit import BurnUnit
from driptorch.personnel import Igniter, IgnitionCrew
from driptorch.firing_techniques._factory import FiringFactory as FiringTechniques
from driptorch.mapping import Map
from driptorch.io import Projector

__all__ = [
    "BurnUnit",
    "Igniter",
    "IgnitionCrew",
    "FiringTechniques",
    "Map",
    "Projector"
]



"""
Folium map wrapper for easy burn unit and animated pattern mapping
"""

# Internal imports
from driptorch.pattern import Pattern
from driptorch.unit import BurnUnit

# External imports
import folium
from folium import plugins


class Map:

    def __init__(self, burn_unit: BurnUnit):
        """Constructor

        Args:
            burn_unit (BurnUnit): Burn unit boundary
        """

        # Configure the folium map object
        self.map = folium.Map(
            location=[0, 0],
            tiles=None
        )

        # Add the basemaps
        self._add_tile_layers()

        # Store the burn unit geojson. We'll add to the map later since
        # we want it plotted on the top
        self.unit_geojson = burn_unit.to_json(
            style={
                'fillColor': '#e8eb34',
                'color': '#0000ff',
                'fillOpacity': 0.0,
                'weight': 4
            }
        )

    def _add_tile_layers(self):
        """Helper method for adding imagery, terrain and street tiles to the map
        """

        # Imagery
        folium.TileLayer(
            tiles='https://basemap.nationalmap.gov/arcgis/rest/services/USGSImageryOnly/MapServer/tile/{z}/{y}/{x}',
            attr='Tiles courtesy of the <a href="https://usgs.gov/">U.S. Geological Survey</a>',
            name='Imagery',
        ).add_to(self.map)

        # Hybrid (imagery + contours)
        folium.TileLayer(
            tiles='https://basemap.nationalmap.gov/arcgis/rest/services/USGSImageryTopo/MapServer/tile/{z}/{y}/{x}',
            attr='Tiles courtesy of the <a href="https://usgs.gov/">U.S. Geological Survey</a>',
            name='Contours'
        ).add_to(self.map)

        # OpenStreetMap
        folium.TileLayer('openstreetmap', name='Street').add_to(self.map)

    def add_firing_area(self, firing_area: BurnUnit):
        """Add the interior firing area to the map as a yellow polygon

        Args:
            firing_area (BurnUnit): Interior firing area
        """

        # Write to GeoJSON and style
        geojson = firing_area.to_json(
            style={
                'fillColor': '#ffff00',
                'color': '#ffff00',
                'fillOpacity': 0.1
            }
        )

        # Add it to the folium map
        folium.GeoJson(
            data=geojson,
            style_function=Map.styling,
            name='Firing Area'
        ).add_to(self.map)

    def add_blackline_area(self, blackline_area: BurnUnit):
        """Add the blackline area to the map as a black polygon

        Args:
            blackline_area (BurnUnit): Dlackline area (difference between
                the burn unit and firing area)
        """

        # Write to GeoJSON and style
        geojson = blackline_area.to_json(
            style={
                'fillColor': '#000000',
                'color': '#000000',
                'fillOpacity': 0.5,
                'weight': 0
            }
        )

        # Add it to the map
        folium.GeoJson(
            data=geojson,
            style_function=Map.styling,
            name='Blackline Area',
        ).add_to(self.map)

    def add_pattern(self, pattern: Pattern):
        """Add the pattern to the map as an animated timedstamped
        GeoJSON. Lines, dashes and dots are red.

        Args:
            pattern (Pattern): Ignition patterns
        """

        # Add the pattern to the map
        plugins.TimestampedGeoJson(
            pattern.to_json(),
            transition_time=10,
            add_last_point=False,
            period='PT1S'
        ).add_to(self.map)

    def show(self) -> folium.Map:
        """Method to show the map in a notebook

        Returns:
            folium.Map: Folium map object
        """

        # Add the burn unit boundary to the map
        unit_layer = folium.GeoJson(
            data=self.unit_geojson,
            style_function=Map.styling,
            name='Burn Unit',
        ).add_to(self.map)

        # Fit the map to the burn unit
        self.map.fit_bounds(unit_layer.get_bounds())

        # Add the layer control button
        folium.LayerControl().add_to(self.map)

        return self.map

    @staticmethod
    def styling(feature: dict):
        """Helper method for parsing styles from the GeoJSON features

        Args:
            feature (dict): GeoJSON feature

        Returns:
            dict: Styling properties
        """
        return {
            'fillColor': feature['properties']['fillColor'],
            'color': feature['properties']['color'],
            'weight': feature['properties'].get('weight', 2),
            'fillOpacity': feature['properties'].get('fillOpacity', 0.1)
        }



"""
DripTorch I/O helper functions
"""

# Internal imports
from driptorch.templates import quicfire
from driptorch.errors import *

# External imports
from folium import Polygon
import numpy as np
import pyproj
from shapely.geometry import mapping, shape, MultiLineString, LineString, Point, MultiPoint
from shapely.geometry.base import BaseGeometry
from shapely.ops import transform


class Projector:
    """
    Helper class to handle reprojections during I/O operations.
    """

    def __init__(self, src_epsg, dst_epsg):
        """Constructor will initialize a function for forward projection and
        a function for backwards projection.
        """

        # Configure transformer for forward projections
        self.forward_proj = pyproj.Transformer.from_proj(
            pyproj.Proj(f'epsg:{src_epsg}'),
            pyproj.Proj(f'epsg:{dst_epsg}'),
            always_xy=True
        )

        # Configure transform for inverse projections
        self.backward_proj = pyproj.Transformer.from_proj(
            pyproj.Proj(f'epsg:{dst_epsg}'),
            pyproj.Proj(f'epsg:{src_epsg}'),
            always_xy=True
        )

    def forward(self, geometry: BaseGeometry) -> BaseGeometry:
        """Project from source EPSG to destination EPSG

        Args:
            geometry (BaseGeometry): Input geometry to reproject

        Returns:
            BaseGeometry: Reprojected geometry
        """

        return transform(self.forward_proj.transform, geometry)

    def backward(self, geometry: BaseGeometry) -> BaseGeometry:
        """Project from destimation EPSG to source EPSG

        Args:
            geometry (BaseGeometry): Input geometry to reproject

        Returns:
            BaseGeometry: Reprojected geometry
        """

        return transform(self.backward_proj.transform, geometry)

    @staticmethod
    def estimate_utm_epsg(lon, lat):
        return int(32700-round((45+lat)/90, 0)*100+round((183+lon)/6, 0))

    @classmethod
    def web_mercator_to_utm(cls, geometry: BaseGeometry) -> BaseGeometry:

        lon, lat = list(geometry.centroid.coords[0])
        utm_epsg = cls.estimate_utm_epsg(lon, lat)

        projector = cls(4326, utm_epsg)

        return utm_epsg, projector.forward(geometry)

    @classmethod
    def to_web_mercator(cls, geometry: BaseGeometry | dict, src_epsg: int) -> BaseGeometry | dict:
        """Convenience method to project a shapely geometry or GeoJSON feature to web mercator

        Args:
            geometry (BaseGeometry | dict): Either a shapely geometry or GeoJSON
                feature (not a feature collection).
            src_epsg (int): EPSG code of the CRS that the spatial data are currently projected in.

        Returns:
            BaseGeometry | dict: A shapely geometry or GeoJSON feature projected in 4326
        """

        projector = cls(src_epsg, 4326)

        if isinstance(geometry, dict):
            geometry = shape(geometry)
            return mapping(projector.forward(geometry))

        return projector.forward(geometry)


def read_geojson_polygon(geojson: dict) -> Polygon:
    """Parse a GeoJSON to a shapely Polygon

    Args:
        geojson (dict): Input GeoJSON dictionary projected in 4326

    Raises:
        GeojsonError: Raise error if we can't figure out the formatting

    Returns:
        Polygon: Shapely polygon geometry
    """

    # If the geojson is a feature collection then we loop over the features
    # and find the first instance of a polygon geometry type
    if geojson['type'] == 'FeatureCollection':
        for feature in geojson['features']:
            if feature['geometry']['type'].lower() == 'polygon':
                geometry = shape(feature['geometry'])
                break

    # Maybe it's just the geometry?
    elif geojson['type'].lower() == 'Polygon':
        geometry = shape(geojson)

    # Fix your shit, we're not gonna to keep trying to guess
    else:
        raise GeojsonError(GeojsonError.read_error)

    return geometry


def write_geojson(geometries: list[BaseGeometry], utm_epsg: int, properties={}, style={}) -> dict:
    """Write a list of shapely geometries to GeoJSON

    Args:
        geometries (list[BaseGeometry]): List of shapely geometries
        properties (dict, optional): Properties for each feature. Defaults to {}.
        style (dict, optional): Rendering style applied to all features. Defaults to {}.

    Returns:
        dict: GeoJSON
    """

    # Get a projector instance for inverse projection
    projector = Projector(utm_epsg, 4326)

    # Get the names of all the props
    property_names = properties.keys()

    # Loop over each geometry in the input list and write to GeoJSON
    features = []
    for i, geometry in enumerate(geometries):

        props = {}
        for name in property_names:
            props[name] = properties[name][i]

        features.append(
            {
                'type': 'Feature',
                'properties': props | style,
                'geometry': mapping(projector.forward(geometry))
            }
        )

    # Compile the features in a feature collection
    geojson = {
        'type': 'FeatureCollection',
        'features': features
    }

    return geojson


def write_quicfire(geometry: list, times: list) -> str:
    """Writes a QUIC-fire ignition file

    Args:
        geometry (list): List of geometry to write
        times (list): Arrival times corresponding to the geometry coordinates (seconds)

    Raises:
        ExportError: Error if Point and (Multi)LineString geometry types are mixed

    Returns:
        str: QUIC-fire formated ignition file
    """

    rows = ''
    n_rows = 0

    # TODO: #20 Cleanup QF export method
    if all(isinstance(x, (LineString, MultiLineString)) for x in geometry):
        for i, geom in enumerate(geometry):
            time = times[i]
            if isinstance(geom, LineString):
                geom = [geom]
                time = [time]
            for j, part in enumerate(geom):
                coords = np.array(part.coords)
                t = time[j]
                for k, xy in enumerate(coords[:-1]):
                    rows += f'{xy[0]} {xy[1]} {coords[k+1,0]} {coords[k+1,1]} {t[k]} {t[1]}\n'
                    n_rows += 1
        file = quicfire.fmt_5.substitute(n_rows=n_rows, rows=rows)

    elif all(isinstance(x, (Point, MultiPoint)) for x in geometry):
        for i, geom in enumerate(geometry):
            time = times[i]
            if isinstance(geom, Point):
                geom = [geom]
                time = [time]
            for j, part in enumerate(geom):
                xy = np.array(part.coords[0])
                rows += f'{xy[0]} {xy[1]} {time[j]}\n'
                n_rows += 1
        file = quicfire.fmt_4.substitute(n_rows=n_rows, rows=rows)

    else:
        raise ExportError(ExportError.incompatible_line_types)

    return file



import pathlib
from setuptools import setup

# --------------------------------------------------------
# This is commented out due to issues with conda skeleton,
# will need to find another way to pull in git tags
# --------------------------------------------------------
# import subprocess

# # Fetch version number from git tag
# version = (
#     subprocess.run(['git', 'describe', '--tags'], stdout=subprocess.PIPE)
#     .stdout.decode('utf-8')
#     .strip()
# )

# # Remove the 'v' prefix if present
# if version[0] == 'v':
#     version = version[1:]

version = '0.3.0'

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name='driptorch',
    packages=['driptorch'],
    version=version,
    license='MIT',
    description='Ignition pattern simulator for prescribed firing techniques',
    long_description=README,
    long_description_content_type="text/markdown",
    author='Holtz TDS',
    author_email='lucas@holtztds.com',
    url='https://github.com/teamholtz/driptorch',
    keywords=['ignition pattern', 'fire',
              'firing techniques', 'prescribed burn'],
    install_requires=[
        'Shapely==1.8.2',
        'awkward==1.8.0',
        'numpy==1.22.4',
        'pandas==1.4.2',
        'pyproj==3.3.1',
        'folium==0.12.1.post1'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
)
